/*	$Id: RioArchive.cpp,v 1.9 2001/12/15 14:18:56 kent Exp $	-*- c++ -*- */
/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>, TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "RioArchive.h"

#if 0
# define DPR(a) dprintf a
#else
# define DPR(a)
#endif

void
dprintf(char const *format, ...)
{
	va_list ap;
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
}


class ClassDescriptor {
	char const *name;
	bool staticp;

public:
        ClassDescriptor(const char* name, bool s) {
                this->staticp = s;
                if (s) {
                        this->name = name;
                } else {
                        this->name = strdup(name);
                }
        }

        virtual ~ClassDescriptor() {
                if (!this->staticp)
                        free((void*)this->name);
        }

        /**
         * ClassDescriptor::load() may return NULL.
         * At this moment, the return value must be an ObjectOcean instance or NULL.
         */
        virtual RioObject* load(RioArchive* rio, uint16_t ver, uint16_t aux) const;

        const char* getName() const {
                return this->name;
        }

        bool isStatic() const {
                return this->staticp;
        }
};

/* ================================================================
 * ClassMap
 * ================================================================ */

class ClassMap {
	uint16_t id;
        uint16_t version;
	ClassMap* next;
	ClassDescriptor const *desc;

public:
        ClassMap(uint16_t id, uint16_t v, const ClassDescriptor* desc, ClassMap* next) {
                this->id = id;
                this->version = v;
                this->desc = desc;
                this->next = next;
        }

        ~ClassMap() {
                delete this->next;
                if (this->desc && !this->desc->isStatic())
                        delete const_cast<ClassDescriptor*>(this->desc);
        }

        uint16_t getId() const {
                return this->id;
        }
        uint16_t getVersion() const {
                return this->version;
        }
        const ClassDescriptor* getClass() const {
                return this->desc;
        }

        ClassMap* searchFor(uint16_t id) {
                ClassMap* map = this;
                while (map != NULL) {
                        if (map->id == id)
                                return map;
                        map = map->next;
                }
                return NULL;
        }
};

/* ================================================================
 * ObjectOceanNode
 * ================================================================ */

ObjectOceanNode::ObjectOceanNode(RioArchive* rio)
{
        int len;
        uint16_t aux;
        const ClassMap* classMap;

        this->robj = NULL;
        this->desc = NULL;
        this->name = NULL;
        this->parent = NULL;

        len = rio->get8();
        this->name = (char*)malloc(len+1);
        rio->read((uint8_t*)this->name, len);
        this->name[len] = 0;
        rio->seek(3);           /* skip: 08 C0 00 */

        classMap = rio->loadClass();
        this->begin = rio->get32le();
        this->size = rio->get32le();
        if (this->size == 0)
                this->begin = 0; // Avoid offset-collision
        aux = rio->get16le();
        if (classMap) {
                this->desc = classMap->getClass();
                DPR(("name=\"%s\", class=\"%s\", begin=0x%08X, size=0x%08X\n",
                     this->name, this->desc->getName(),
                     this->begin, this->size));
                this->robj = this->desc->load(rio, classMap->getVersion(), aux);
                if (this->robj != NULL)
                        this->robj->setParent(this);
        } else {
                free(this->name);
                this->name = NULL;
        }
}

ObjectOceanNode::~ObjectOceanNode()
{
        DPR(("ObjectOceanNode::~ObjectOceanNode()\n"));
        delete this->robj;
        free(this->name);
}

char pwd[1024], *tail=pwd; /* XXX */
void
ObjectOceanNode::dump() {
        char *saved_tail;
        if (!this)
                return;
        saved_tail = tail;
        printf("%s%s:\n\tclass=\"%s\", begin=0x%08X, size=0x%08X\n",
               pwd, this->name,
               this->desc->getName(), this->begin, this->size);
        sprintf(tail, "%s/", this->name);
        tail+=strlen(this->name)+1;
        if (this->robj)
                this->robj->dump();
        tail=saved_tail;
        *tail = 0;
}

int
ObjectOceanNode::getTotalNum(const NameFilter* filter) const
{
        if (this->size == 0)
                return 0;
        if (this->hasChildren()) {
                ObjectOcean* oo = (ObjectOcean*)this->robj;
                return oo->getTotalNum(filter);
        }
        if (filter && !filter->accept(this->name))
                return 0;
        return 1;
}

int
ObjectOceanNode::storePath(char* path, const int len) const
{
        int namelen;
        int pos = 0;

        if (this->parent != NULL) {
                pos = this->parent->storePath(path, len);
                if (pos+2 >= len)
                        return pos;
                path[pos++] = '\\';
                path[pos] = 0;
        }
        namelen = strlen(this->name);
        if (pos+namelen+1 >= len)
                return pos;
        strcpy(path+pos, this->name);
        return pos+namelen;
}

#if _WIN32
#include <windows.h>
#include <winbase.h>
SusieFileInfo*
ObjectOceanNode::store(const NameFilter* filter, SusieFileInfo* sfi) const
{
        if (this->size == 0)
                return sfi;
        if (this->hasChildren()) {
                // qĂꍇ͎qԂāA͕ԂȂB
                ObjectOcean* oo = (ObjectOcean*)this->robj;
                return oo->storeAll(filter, sfi);
        }
        if (filter && !filter->accept(this->name))
                return sfi;

        lstrcpyn((char*)sfi->method, this->desc->getName(), 8);
        sfi->position = this->begin;
        sfi->compsize = this->size;
        sfi->filesize = this->size;
        if (this->parent != NULL) {
                this->parent->storePath(sfi->path, 200);
                strcat(sfi->path, "\\");
        }
        lstrcpyn(sfi->filename, this->name, 200);
        return ++sfi;
}
#endif


/* ================================================================
 * RioObject and ObjectOcean
 * ================================================================ */

RioObject::RioObject(const ClassDescriptor* desc)
{
        this->desc = desc;
}
RioObject::~RioObject()
{
}


ObjectOcean::ObjectOcean(const ClassDescriptor* d, RioArchive* rio, uint16_t num)
                : RioObject(d) {
        this->numObjects = num;
        this->nodes = new ObjectOceanNode*[this->numObjects];
        for (int i = 0; i < this->numObjects; i++) {
                this->nodes[i] = NULL;
        }
        for (int j = 0; j < this->numObjects; j++) {
                this->nodes[j] = new ObjectOceanNode(rio);
        }
}

ObjectOcean::~ObjectOcean()
{
        DPR(("ObjectOcean::~ObjectOcean()\n"));
        for (int i = 0; i < this->numObjects; i++) {
                delete this->nodes[i];
        }
        delete[] this->nodes;
}

void
ObjectOcean::dump() const
{
        if (this->numObjects)
                printf("// class=%s\tnum objects=%d\n",
                       this->desc->getName(), this->numObjects);
        for (int i = 0; i < this->numObjects; i++) {
                this->nodes[i]->dump(); // ObjectOceanNode::dump()
        }
}

void
ObjectOcean::setParent(ObjectOceanNode* parent)
{
        for (int i = 0; i < this->numObjects; i++)
                this->nodes[i]->setParent(parent);
}

int
ObjectOcean::getTotalNum(const NameFilter* filter) const
{
        int total = 0;
        for (int i = 0; i < this->numObjects; i++)
                total += this->nodes[i]->getTotalNum(filter);
        return total;
}


static const char*
mbschr(const char* str, int ch)
{
        const char* p = str;
        while (*p) {
                if (*p == ch)
                        return p;
                if (IsDBCSLeadByte(*p))
                        p++;
                p++;
        }
        return NULL;
}

const ObjectOceanNode*
ObjectOcean::searchFor(const char* name, bool ignoreCase) const
{
        int len = strlen(name);
        const char* sep = mbschr(name, '\\');
        if (sep == NULL)  sep = mbschr(name, '/');
        if (sep == NULL)  sep = name+len;

        char* part = (char*)alloca(sep - name + 1);
        strncpy(part, name, sep-name);
        part[sep-name] = 0;

        for (int i = 0;  i < this->numObjects;  i++) {
                const ObjectOceanNode* node = this->nodes[i];
                int comp = ignoreCase ? strcasecmp(node->getName(), part)
                        : strcmp(node->getName(), part);
                if (comp == 0) {
                        if (*sep == 0)
                                return node->hasChildren() ? NULL : node;
                        if (node->hasChildren()) {
                                ObjectOcean* oo = (ObjectOcean*)node->getObject();
                                node = oo->searchFor(sep+1, ignoreCase);
                                if (node != NULL)
                                        return node;
                        }
                }
        }
        return NULL;
}

const ObjectOceanNode*
ObjectOcean::searchFor(uint32_t offset) const
{
        for (int i = 0;  i < this->numObjects;  i++) {
                const ObjectOceanNode* node = this->nodes[i];
                if (node->hasChildren()) {
                        ObjectOcean* oo = (ObjectOcean*)node->getObject();
                        node = oo->searchFor(offset);
                        if (node != NULL)
                                return node;
                } else
                        if (node->getOffset() == offset)
                                return node;
        }
        return NULL;
}

#ifdef _WIN32
SusieFileInfo*
ObjectOcean::storeAll(const NameFilter* filter, SusieFileInfo* sfi) const
{
        for (int i = 0;  i < this->numObjects;  i++)
                sfi = this->nodes[i]->store(filter, sfi);
        return sfi;
}
#endif


/* ================================================================
 * ClassDescriptor
 * ================================================================ */

RioObject*
ClassDescriptor::load(RioArchive* rio, uint16_t ver, uint16_t aux) const {
        if (aux > 0)
                return new ObjectOcean(this, rio, aux);
        else
                return NULL;
}


class VmFuncClass: public ClassDescriptor {
public:
        VmFuncClass(): ClassDescriptor("CVmFunc", true) {
        }

        RioObject* load(RioArchive* rio, uint16_t ver, uint16_t aux) const {
                DPR(("VmFuncClass::load() ver=0x%4.4x aux=0x%4.4x\n", ver, aux));
                if (ver <= 2) {
                        return ClassDescriptor::load(rio, ver, aux);
                } else {
                        int i;
                        int len;
                        ClassMap const *classmap;
                        uint16_t v;

                        /* XXX */
                        for (i=aux; i>0; i--) {
                                uint8_t buf[256];
                                uint8_t b0, b1, b2;
                                len = rio->get8();
                                rio->read(buf, len);
                                buf[len] = 0;
                                b0 = rio->get8();
                                b1 = rio->get8();
                                b2 = rio->get8();
                                DPR((" vmfunc: '%s' %2.2x %2.2x %2.2x\n",
                                     buf, b0, b1, b2));
                                if (b0 != 0x08) {
                                        uint16_t p0;
                                        uint8_t b3, b4, b5;
                                        p0 = rio->get16le();
                                        b3 = rio->get8();
                                        b4 = rio->get8();
                                        b5 = rio->get8();
                                }
                                classmap = rio->loadClass();
                                if (classmap == NULL) {
                                        rio->seek(5);
                                } if (classmap->getId() == 0) {
                                        len = rio->get16le();
                                        rio->seek(len);
                                        rio->seek(10);
                                } else {
                                	/*p->begin =*/ rio->get32le();
                                	/*p->size =*/ rio->get32le();
                                        v = rio->get16le();
                                        if (v) {
                                                DPR(("warn: vm_func : aux=%u\n", v));
                                        }
                                }
                        }
                        return NULL;
                }
        }
};


/* ================================================================
 * RioArchive
 * ================================================================ */

RioArchive::RioArchive(Stream* st) {
        this->stream = st;
        this->top = NULL;
        this->currentCount = 1;
        this->cdesc[0] = new ClassDescriptor("CObjectOcean", true);
        this->cdesc[1] = new VmFuncClass();
        this->cdesc[2] = NULL;
        this->nullClassMap = new ClassMap(0, 0, NULL, NULL);
        this->topObject = NULL;
}


RioArchive::~RioArchive() {
        DPR(("RioArchvie::~RioArchive()\n"));
        delete this->topObject;
        delete this->top;
        delete const_cast<ClassDescriptor*>(this->cdesc[0]);
        delete const_cast<ClassDescriptor*>(this->cdesc[1]);
        delete const_cast<ClassMap*>(this->nullClassMap);
        delete this->stream;
}


const ClassDescriptor*
RioArchive::searchStaticsFor(const char* cname)
{
	ClassDescriptor const * const *cdpp;
        for (cdpp = this->cdesc; *cdpp; cdpp++) {
                if (!strcmp(cname, (*cdpp)->getName()))
                        break;
        }
        return *cdpp;
}


#define ID_NEW_CLASS_TAG	0xffff
#define ID_SMALL_CLASS_TAG	0x8000
ClassMap const *
RioArchive::loadClass()
{
	uint16_t tag, len;
	ClassMap const *classmap = NULL;
	ClassDescriptor const * cdp;
	char buf[256];

        DPR(("RioArchive::loadClass()\n"));
        this->currentCount++;
	tag = this->get16le();
        if (tag == 0x0010)
                tag = this->get16le();
        else if (tag == 0x0011) {
                tag = this->get16le();
                if (tag == 0x0001)
                        tag = this->get16le();
#if 0
        } else if (tag == 0x1800) { // Q.rio L
                tag = this->get16le();
#endif
        }
	if (tag == ID_NEW_CLASS_TAG) {
		/* new class */
		uint16_t ver = this->get16le(); /* schema */
		len = this->get16le();
                this->read((uint8_t*)buf, len);
		buf[len]=0;
                cdp = this->searchStaticsFor(buf);
		if (!cdp) {
			DPR(("warn: unknown class (name=\"%s\")\n", buf));
			cdp = new ClassDescriptor(buf, false);
		}
                DPR(("class='%s' version=0x%4.4x\n", cdp->getName(), ver));
		this->top = new ClassMap(this->currentCount++, ver, cdp, this->top);
		classmap = this->top;
	} else if ((tag&ID_SMALL_CLASS_TAG) == ID_SMALL_CLASS_TAG) {
		/* cached class */
		tag &= ~ID_SMALL_CLASS_TAG;
                classmap = this->top->searchFor(tag);
		if (!classmap) {
			DPR(("warn: cannot found cached class (ID=0x%04X)\n",
			     tag));
		}
	} else if (tag==0) {
		classmap = this->nullClassMap;
	} else {
		DPR(("warn: illegal tag (0x%04X)\n", tag));
                classmap = NULL;
	}
	return classmap;
}


#define ID_RIO		0x596E32CD
RioObject*
RioArchive::load()
{
	uint32_t magic;
	uint16_t aux;
        const ClassMap* classMap;

        DPR(("%s: start\n", __func__));
	magic = this->get32le();
	if (magic != ID_RIO)
		return NULL;
	//this->get16le();        /* skip 0x0010 */
        classMap = this->loadClass();
        if (classMap == NULL)
                return NULL;
	this->currentCount++;
	aux = this->get16le();
        if (aux == 0xffff)
                return NULL;
        this->topObject = classMap->getClass()->load(this, classMap->getVersion(), aux);

	return this->topObject;
}


void
RioArchive::extract(const ObjectOceanNode* node, void* buf)
{
        this->stream->seek(node->getOffset(), SEEK_SET);
        this->stream->read((uint8_t*)buf, node->getSize());
}

