/*	$Id: StreamImpl.cpp,v 1.4 2001/10/05 01:16:03 kent Exp $ -*- c++ -*-	*/
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "StreamImpl.h"


FileStream::FileStream(FILE* fp) {
        this->fp = fp;
        this->offset = 0;
        fseek(this->fp, 0, SEEK_SET);
}

FileStream::FileStream(FILE* fp, long offset) {
        this->fp = fp;
        this->offset = offset;
        fseek(this->fp, this->offset, SEEK_SET);
}

FileStream::~FileStream() {
        fclose(this->fp);
}

int FileStream::get() {
        return fgetc(this->fp);
}

bool FileStream::isEof() {
        return feof(this->fp);
}

void FileStream::seek(long off, int whence) {
        if (whence == SEEK_SET)
                fseek(this->fp, this->offset + off, whence);
        else
                fseek(this->fp, off, whence);
}


MemoryStream::MemoryStream(const uint8_t* buffer, size_t size) {
        this->buffer = buffer;
        this->size = size;
        this->offset = 0;
}

int MemoryStream::get() {
        if (this->offset >= this->size) {
                return EOF;
        }
        return this->buffer[this->offset++];
}

bool MemoryStream::isEof() {
        return this->offset >= this->size;
}

void MemoryStream::seek(long off, int whence) {
        switch (whence) {
        case SEEK_CUR:
                this->offset += off;
                break;
        case SEEK_SET:
                this->offset = off;
                break;
        case SEEK_END:
                this->offset = this->size - off;
        }
}


#ifdef _WIN32
Win32FileStream::Win32FileStream(HANDLE fd, long offset) {
        this->eof = false;
        SetFilePointer(fd, offset, NULL, FILE_BEGIN);
        this->offset = offset;
        this->fd = fd;
        this->fillBuffer();
}

Win32FileStream::~Win32FileStream() {
        CloseHandle(this->fd);
}

void Win32FileStream::fillBuffer() {
        int message(const char* mes, int v);
        this->reader = 0;
        this->size = 0;
        DWORD len;
        while (this->size < W32FS_BUFFER_MAX
               && ReadFile(this->fd, &this->buffer[this->size],
                           W32FS_BUFFER_MAX - this->size, &len, NULL)
               && len > 0)
                this->size += len;
}

int Win32FileStream::get() {
        if (this->eof)
                return EOF;
        if (this->reader >= this->size) {
                this->fillBuffer();
                if (this->size <= 0) {
                        this->eof = true;
                        return EOF;
                }
        }
        return this->buffer[this->reader++];
}

int Win32FileStream::read(uint8_t* buffer, size_t length) {
        size_t done = 0;
        DWORD len;
        if (this->reader < this->size) {
                done = length < this->size - this->reader ? length : this->size - this->reader;
                memcpy(buffer, this->buffer + this->reader, done);
                this->reader += done;
        }
        while (done < length
               && ReadFile(this->fd, buffer + done, length - done, &len, NULL)
               && len > 0)
                done += len;
        return done;
}

bool Win32FileStream::isEof() {
        return this->eof;
}

void Win32FileStream::seek(long off, int whence) {
        switch (whence) {
        case SEEK_CUR:
                if (off >= 0) {
                        while (off-- > 0)
                                this->get();
                } else {
                        SetFilePointer(this->fd, off - (this->size - this->reader), NULL, FILE_CURRENT);
                        this->fillBuffer();
                }
                break;
        case SEEK_SET:
                SetFilePointer(this->fd, off + this->offset, NULL, FILE_BEGIN);
                this->fillBuffer();
                break;
        case SEEK_END:
                SetFilePointer(this->fd, off, NULL, FILE_END);
                this->fillBuffer();
                break;
        }
}
#endif // _WIN32
